﻿-----------------------------------------------------------------------
-- AddOn namespace.
-----------------------------------------------------------------------
local LibStub = _G.LibStub
local WardrobeAchiever = LibStub("AceAddon-3.0"):NewAddon("WardrobeAchiever")
local AL = LibStub("AceLocale-3.0"):GetLocale("WardrobeAchiever", false)

local ADDON_NAME, private = ...

local DEBUG_MODE = true;

local classID = nil
local faction = nil
local achievementShowedId = nil
local playerID = nil

local ACHIEVEMENTUI_WIDTH = 500;
local SLIDERUI_WIDTH = 18;
local COLLECTIONUI_WIDTH = 430;
local ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT = 84;

local ADDON_NAME_TEXT = "WardrobeAchiever"
local NUM_ACHIEVEMENTS = table.getn(private.ACHIEVEMENTS)
local SCROLL_MAX_VALUE = 0
local SCROLL_STEP_VALUE = ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT

-- Settings
local PROFILE_DEFAULTS = {
	char = {
		general = {
			hideCompleted = false
		},
		minimap = {
			hide = false,
		},
	},
	global = {
		stats = {
			achi = {
			
			}
		},
		characterInfo = {
		
		}
	}
}

---------------------------
-- Minimap icon          --
---------------------------
local MapIcon = LibStub("LibDBIcon-1.0")
local wardroveAchieverLDB = LibStub("LibDataBroker-1.1"):NewDataObject(ADDON_NAME_TEXT, {
	type = "data source",
	text = ADDON_NAME_TEXT,
	label = ADDON_NAME_TEXT,
	icon = "Interface\\Icons\\INV_Chest_Cloth_61",
	OnClick = function(self, button)
		if button == "RightButton" then
			InterfaceOptionsFrame_OpenToCategory("WardrobeAchiever")
		else
			WardrobeAchiever:Toggle()
		end
	end,
	OnTooltipShow = function(tt)
		tt:AddLine(ADDON_NAME_TEXT)
		tt:AddLine("|cffffffff" .. AL["MINIMAP_LEFT_CLICK"])
		tt:AddLine("|cffffffff" .. AL["MINIMAP_RIGHT_CLICK"])
	end,
})

-- --------------------- --
-- Addon frame
-- --------------------- --
local wardroveAchieverFrame = CreateFrame("Frame", "wardroveAchieverFrame", UIParent, "UIPanelDialogTemplate") 
wardroveAchieverFrame:SetSize(ACHIEVEMENTUI_WIDTH + SLIDERUI_WIDTH + COLLECTIONUI_WIDTH, 600)
wardroveAchieverFrame:SetPoint("CENTER")
wardroveAchieverFrame:SetMovable(true)
wardroveAchieverFrame:EnableMouse(true)
wardroveAchieverFrame.Title:SetText("WardrobeAchiever")
wardroveAchieverFrame:RegisterForDrag("LeftButton")
wardroveAchieverFrame:SetScript("OnDragStart", wardroveAchieverFrame.StartMoving)
wardroveAchieverFrame:SetScript("OnDragStop", wardroveAchieverFrame.StopMovingOrSizing)
wardroveAchieverFrame:Hide()
wardroveAchieverFrame:SetScript("OnHide", function(self)
	WardrobeAchiever:ApplyFilters()
end)
wardroveAchieverFrame:SetScript("OnShow", function(self)
	-- Forze selecting first if nothing selected (shouldnt ocurr)
	if (not wardroveAchieverFrame.collectionsFrame.selectedSetId) then
		WardrobeAchiever:WA_selectFirst()
	end
end)
wardroveAchieverFrame:RegisterEvent("ZONE_CHANGED_NEW_AREA");
wardroveAchieverFrame:RegisterEvent("TRANSMOG_COLLECTION_UPDATED");
wardroveAchieverFrame:SetScript("OnEvent", function(self, event, ...)
	if (event == "TRANSMOG_COLLECTION_UPDATED") then
		-- Forze load panels
		WardrobeAchiever:WA_LoadPanels()
		
		-- Selects items earned
		wardroveAchieverFrame.collectionsFrame:SelectAchievementCollectionUpdated()
		self:UnregisterEvent("TRANSMOG_COLLECTION_UPDATED");
	elseif (event == "ZONE_CHANGED_NEW_AREA") then
		-- If not in dungeon or raid ignore
		local inInstance, instanceType = IsInInstance()
		if (not inInstance or not instanceType or (instanceType ~= "party" and instanceType ~= "raid")) then
			return
		end
	
		-- Select achievement in the new zone
		local mapID, _ = GetCurrentMapAreaID()
		
		-- Forze load panels
		WardrobeAchiever:WA_LoadPanels()
		
		if (achievementsFrame.buttons) then
			local zoneFound = nil
			for i=1, #achievementsFrame.buttons do
				local button = achievementsFrame.buttons[i]
				if (private.ACHIVE_ZONES[button.id]) then
					for j=1, #private.ACHIVE_ZONES[button.id] do
						local zoneId = private.ACHIVE_ZONES[button.id][j]
						if (zoneId == mapID) then
							if (not button.completed) then
								print("|cFF00FF00[WardrobeAchiever]: |cFFFFFFFF"..AL["ZONE_ACHIEVEMENT_FOUND"])
								button:Click();
								WardrobeAchiever:WA_moveScrollerTo(button)
							end
							
							zoneFound = true
							break
						end
					end
				end
				
				if (zoneFound) then
					break
				end
			end
		end
	end
end)

tinsert(UISpecialFrames, wardroveAchieverFrame:GetName())

-- --------------------- --
-- Achievements frame
-- --------------------- --
local scrollFrame = _G.CreateFrame("ScrollFrame", "scrollFrame", wardroveAchieverFrame)
scrollFrame:SetClipsChildren(true)
scrollFrame:SetPoint("TOPLEFT", 8, -26)
scrollFrame:SetPoint("BOTTOMRIGHT", wardroveAchieverFrame, "BOTTOMLEFT", ACHIEVEMENTUI_WIDTH + 8, 10)
scrollFrame:EnableMouseWheel(true)
scrollFrame:SetScript("OnMouseWheel", function(self, delta)
	local current = self.scrollbar:GetValue()

	if (current) then
		if (delta < 0) and (current < SCROLL_MAX_VALUE) then
			self.scrollbar:SetValue(current + SCROLL_STEP_VALUE)
		elseif (delta > 0) and (current > 1) then
			self.scrollbar:SetValue(current - SCROLL_STEP_VALUE)
		end
	end
end)

scrollFrame.scrollbar = CreateFrame("Slider", nil, wardroveAchieverFrame, "UIPanelScrollBarTemplate") 
scrollFrame.scrollbar:SetPoint("TOPLEFT", scrollFrame, "TOPRIGHT", -1, -16) 
scrollFrame.scrollbar:SetPoint("BOTTOMLEFT", scrollFrame, "BOTTOMRIGHT", -1, 15) 
scrollFrame.scrollbar:SetValueStep(SCROLL_STEP_VALUE) 
scrollFrame.scrollbar.scrollStep = SCROLL_STEP_VALUE
scrollFrame.scrollbar:SetValue(1) 
scrollFrame.scrollbar:SetWidth(SLIDERUI_WIDTH) 
scrollFrame.scrollbar:SetScript("OnValueChanged", function (self, value) 
	scrollFrame:SetVerticalScroll(value) 
end) 

local achievementsFrame = CreateFrame("Frame", "achievementsFrame", scrollFrame)
achievementsFrame:SetSize(ACHIEVEMENTUI_WIDTH, ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT * NUM_ACHIEVEMENTS)
achievementsFrame:SetPoint("TOPLEFT", 0, 0)
achievementsFrame:RegisterEvent("ACHIEVEMENT_EARNED");
achievementsFrame:SetScript("OnEvent", function(self, event, achiID)
	if (event == "ACHIEVEMENT_EARNED") then
		-- Forze load panels
		WardrobeAchiever:WA_LoadPanels()
		
		if (achievementsFrame.buttons) then
			table.foreach(achievementsFrame.buttons, function(index, button)
				if (button.id == achiID) then
					button:Update();
					button:Click();
					WardrobeAchiever:WA_moveScrollerTo(button)
				end
			end)
		end
	end
end)
scrollFrame:SetScrollChild(achievementsFrame);

-- ------------------------------ --
-- Collections frame - background
-- ----------------------------- --
wardroveAchieverFrame.collectionsFrame = CreateFrame("Frame", "collectionsFrame", wardroveAchieverFrame, "InsetFrameTemplate")
wardroveAchieverFrame.collectionsFrame:SetPoint("TOPLEFT", scrollFrame.scrollbar, "TOPRIGHT", 0, 16)
wardroveAchieverFrame.collectionsFrame:SetPoint("BOTTOMRIGHT", wardroveAchieverFrame, "BOTTOMRIGHT", -5, 10)

-- --------------------- --
-- Initialize
-- --------------------- --
function WardrobeAchiever:OnInitialize()
	local class, classFileName, internalClassID = UnitClass("player");
	classID = internalClassID
	local internalFaction, _ = UnitFactionGroup("player");
	faction = internalFaction
	playerID = GetUnitName("player", false).."-"..GetRealmName()
	
		-- Adds about panel to wow options
	local about_panel = LibStub:GetLibrary("LibAboutPanel", true)
	if about_panel then
		self.optionsFrame = about_panel.new(nil, "WardrobeAchiever")
	end
	
	-- Load savedvariables
	self.db = LibStub("AceDB-3.0"):New("WardrobeAchieverDB", PROFILE_DEFAULTS, true)
	self:SetupOptions()
	private.db = self.db.char
	private.global = self.db.global
	
	if (not private.global.characterInfo[playerID]) then
		private.global.characterInfo[playerID] = { class = class, classFileName = classFileName }
	end
	
	-- Initialize minimap
	MapIcon:Register("WardrobeAchiever", wardroveAchieverLDB, private.db.minimap)
end

function WardrobeAchiever:getAchiSetID(achiID)
	if (faction == 'Horde' or faction == 'Neutral') then
		return private.ACHIVE_BASESET[achiID][classID].horde
	else
		return private.ACHIVE_BASESET[achiID][classID].alliance
	end
end

function WardrobeAchiever:getSetAchiID(baseSetID)
	local achiIdFound = nil
	if (faction == 'Horde' or faction == 'Neutral') then
		table.foreach(private.ACHIEVEMENTS, function(index, achiID)
			if (private.ACHIVE_BASESET[achiID][classID].horde == baseSetID) then
				achiIdFound = achiID
				return
			end
		end)
	else
		table.foreach(private.ACHIEVEMENTS, function(index, achiID)
			if (private.ACHIVE_BASESET[achiID][classID].alliance == baseSetID) then
				achiIdFound = achiID
				return
			end
		end)
	end
	
	return achiIdFound
end

function WardrobeAchiever:SelectAchiBySetID(baseSetID, variantSetID)
	-- Search for an achievement
	local achiIdFound = WardrobeAchiever:getSetAchiID(baseSetID)
	
	-- If found select it and change the variant set
	if (achiIdFound and achievementsFrame.buttons) then
		table.foreach(achievementsFrame.buttons, function(index, button)
			if (button.id == achiIdFound) then
				button:Click()
				WardrobeAchiever:WA_moveScrollerTo(button)
				
				-- Forze to load variant set
				if (baseSetID ~= variantSetID) then
					wardroveAchieverFrame.collectionsFrame:DisplaySet(variantSetID)
				end
				
				return;
			end
		end)
	end
end


function WardrobeAchiever:ToggleMinimapIcon()
	if (private.db.minimap.hide) then
		MapIcon:Hide("WardrobeAchiever")
	else
		MapIcon:Show("WardrobeAchiever")
	end
end

function WardrobeAchiever:Toggle()
	-- In combat dont do anything
	if InCombatLockdown() then
		return
	end

	-- Initial load of panels
	WardrobeAchiever:WA_LoadPanels()
	
	-- Show addon
	wardroveAchieverFrame:Show()
end

function WardrobeAchiever:WA_LoadPanels()
	-- Initialize panels if not loaded
	if (not achievementsFrame.buttons) then
		achievementsFrame.buttons = {};
		achievementsFrame.achievementSelected = nil
		
		-- Recalculate statistics
		table.foreach(private.ACHIEVEMENTS, function(index, achievementID)
			if ((faction == 'Horde' or faction == 'Neutral') and private.ACHIVE_BASESET[achievementID][classID].horde) then
				WardrobeAchiever:loadStatistics(achievementID, private.ACHIVE_BASESET[achievementID][classID].horde)
			elseif (faction == 'Alliance' and private.ACHIVE_BASESET[achievementID][classID].alliance) then
				WardrobeAchiever:loadStatistics(achievementID, private.ACHIVE_BASESET[achievementID][classID].alliance)
			end
		end)
		
		-- Add achievements
		local currentIndex = 1
		for achievementID, achiStats in spairs(private.global.stats.achi, function (t, a, b) return t[a].percentage > t[b].percentage end) do
			if ((faction == 'Horde' or faction == 'Neutral') and private.ACHIVE_BASESET[achievementID][classID].horde) then
				WardrobeAchiever:addAchievement(achievementsFrame, achievementID, currentIndex);
				currentIndex = currentIndex + 1;
			elseif (faction == 'Alliance' and private.ACHIVE_BASESET[achievementID][classID].alliance) then
				WardrobeAchiever:addAchievement(achievementsFrame, achievementID, currentIndex);
				currentIndex = currentIndex + 1;
			end
		end
		
		-- Initialize collection frame
		WardrobeAchiever:initializeCollectionFrame(wardroveAchieverFrame.collectionsFrame)
		
		-- Applies filters
		WardrobeAchiever:ApplyFilters()
	end
end

function WardrobeAchiever:WA_selectFirst() 
	if (achievementsFrame.buttons) then
		for i=1, #achievementsFrame.buttons do
			if (achievementsFrame.buttons[i]:IsShown()) then
				achievementsFrame.buttons[i]:Click()
				WardrobeAchiever:WA_moveScrollerTo(achievementsFrame.buttons[i]);
				break
			end
		end
	end
end

local previousNumAchivements = nil
function WardrobeAchiever:WA_recalculateScroller(numAchievements) 
	-- Avoid recalculating if there is not changes
	if (previousNumAchivements and previousNumAchivements == numAchievements) then
		return
	else
		previousNumAchivements = numAchievements
		scrollFrame.scrollbar:SetValue(1)
	end

	-- Recalculate slider max value
	SCROLL_MAX_VALUE = ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT * (numAchievements - 1) - scrollFrame:GetHeight()
	if (SCROLL_MAX_VALUE < 0) then
		SCROLL_MAX_VALUE = 1;
	end
	scrollFrame.scrollbar:SetMinMaxValues(1, SCROLL_MAX_VALUE) 
end

function WardrobeAchiever:WA_moveScrollerTo(button) 
	local buttonPositionFound = nil
	local indexTo = 0
	for achievementID, achiStats in spairs(private.global.stats.achi, function (t, a, b) return t[a].percentage > t[b].percentage end) do
		for i=1, #achievementsFrame.buttons do
			if (achievementsFrame.buttons[i].id == achievementID and achievementsFrame.buttons[i]:IsShown()) then
				indexTo = indexTo + 1
				
				if (button.id == achievementID) then
					buttonPositionFound = true
				end
				
				break;
			end
		end
		
		if (buttonPositionFound) then
			break;
		end
	end
	
	local scrollTo = ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT * (indexTo - 1)
	if (scrollTo < 0) then
		scrollTo = 1
	elseif (scrollTo > SCROLL_MAX_VALUE) then
		scrollTo = SCROLL_MAX_VALUE
	end
	scrollFrame.scrollbar:SetValue(scrollTo) 
end

function WardrobeAchiever:ReorderAchievements() 
	local currentIndex = 1
	for achievementID, achiStats in spairs(private.global.stats.achi, function (t, a, b) return t[a].percentage > t[b].percentage end) do
		for i=1, #achievementsFrame.buttons do
			local button = achievementsFrame.buttons[i]
			if (button.id == achievementID) then 
				if (button:IsShown()) then
					button:SetPoint("TOP", 0, -(ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT * (currentIndex - 1)))
					currentIndex = currentIndex + 1;
				end
				
				-- reset tooltip
				if (button.tooltip) then
					button.tooltip:Reset()
				end
				
				break
			end
		end
	end
end

function WardrobeAchiever:ApplyFilters()
	if InCombatLockdown() then
		return
	end
	
	if (achievementsFrame.buttons) then
		local currentIndex = 1
		local needClearSelection = false
		table.foreach(achievementsFrame.buttons, function(index, achievementFrame)
			if (achievementFrame.completed and private.db.general.hideCompleted) then
				if (achievementFrame:IsShown()) then
					achievementFrame:Hide()
					needClearSelection = true
				end
			elseif (achievementFrame.completed and not private.db.general.hideCompleted) then
				if (not achievementFrame:IsShown()) then
					achievementFrame:Show()
					needClearSelection = true
				end
			end
			
			if (achievementFrame:IsShown()) then
				achievementFrame:SetPoint("TOP", 0, -(ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT * (currentIndex - 1)))
				currentIndex = currentIndex + 1
			end
		end)
		
		-- Recalculate scroll child height
		WardrobeAchiever:WA_recalculateScroller(currentIndex) 
		
		-- Reset selection
		if (needClearSelection or not wardroveAchieverFrame.collectionsFrame.selectedSetId) then
			WardrobeAchiever:WA_AchievementFrameAchievements_ClearSelection(achievementsFrame)
			WardrobeAchiever:WA_selectFirst()
		end
	end
end

function WardrobeAchiever:loadStatistics(achievementID, setID, forzeLoad) 
	if (not private.global.stats.achi[achievementID]) then
		private.global.stats.achi[achievementID] = { character = nil, percentage = 0, difficulty = nil, variantSetID = nil, chars = { } }
	end

	if (not private.global.stats.achi[achievementID].chars[playerID] or forzeLoad) then
		private.global.stats.achi[achievementID].chars[playerID] = { piecesCollected = nil, totalPieces = nil, difficulty = nil, variantSetID = nil, percentage = nil }
	end
	
	if (forzeLoad) then
		private.global.stats.achi[achievementID].percentage = 0
	end
	
	local baseSetID = C_TransmogSets.GetBaseSetID(setID);
	local variantSets = WardrobeAchiever:WA_GetVariantSets(baseSetID);

	if (variantSets and next(variantSets)) then
		for i = 1, #variantSets do
			local numSourcesCollected, numSourcesTotal = WardrobeAchiever:WA_GetSetSourceCounts(variantSets[i].setID);
			local newPercentage = round(100 * numSourcesCollected / numSourcesTotal, 0)
			if (not private.global.stats.achi[achievementID].chars[playerID].percentage or private.global.stats.achi[achievementID].chars[playerID].percentage < newPercentage) then
				private.global.stats.achi[achievementID].chars[playerID].piecesCollected = numSourcesCollected
				private.global.stats.achi[achievementID].chars[playerID].totalPieces = numSourcesTotal
				private.global.stats.achi[achievementID].chars[playerID].difficulty = variantSets[i].description
				private.global.stats.achi[achievementID].chars[playerID].variantSetID = variantSets[i].setID
				private.global.stats.achi[achievementID].chars[playerID].percentage = newPercentage
			end
		end
	else
		local numSourcesCollected, numSourcesTotal = WardrobeAchiever:WA_GetSetSourceCounts(setID);
		local newPercentage = round(100 * numSourcesCollected / numSourcesTotal, 0)
		if (not private.global.stats.achi[achievementID].chars[playerID].percentage or private.global.stats.achi[achievementID].chars[playerID].percentage < newPercentage) then
			private.global.stats.achi[achievementID].chars[playerID].piecesCollected = numSourcesCollected
			private.global.stats.achi[achievementID].chars[playerID].totalPieces = numSourcesTotal
			private.global.stats.achi[achievementID].chars[playerID].difficulty = C_TransmogSets.GetSetInfo(setID).description
			private.global.stats.achi[achievementID].chars[playerID].variantSetID = setID
			private.global.stats.achi[achievementID].chars[playerID].percentage = newPercentage
		end
	end
	
	-- Recalculate character with best progress
	for charID, charData in pairs(private.global.stats.achi[achievementID].chars) do
		if (not private.global.stats.achi[achievementID].character or private.global.stats.achi[achievementID].percentage < charData.percentage) then 
			private.global.stats.achi[achievementID].character = charID
			private.global.stats.achi[achievementID].percentage = charData.percentage
			private.global.stats.achi[achievementID].difficulty = charData.difficulty
			private.global.stats.achi[achievementID].variantSetID = charData.variantSetID
			private.global.stats.achi[achievementID].totalPieces = charData.totalPieces
		end
	end
end

function WardrobeAchiever:currentPlayerID()
	return playerID
end

function spairs(t, order)
    -- collect the keys
    local keys = {}
    for k in pairs(t) do keys[#keys+1] = k end

    -- if order function given, sort by it by passing the table and keys a, b,
    -- otherwise just sort the keys 
    if order then
        table.sort(keys, function(a,b) return order(t, a, b) end)
    else
        table.sort(keys)
    end

    -- return the iterator function
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end

function round(number, decimals)
    return tonumber((("%%.%df"):format(decimals)):format(number))
end